@@ -3,11 +3,11 @@
  * async.c
  *	  Asynchronous notification: NOTIFY, LISTEN, UNLISTEN
  *
- * Portions Copyright (c) 1996-2001, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2002, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql-server/src/backend/commands/async.c,v 1.81.2.1 2002/09/30 20:47:22 tgl Exp $
+ *	  $Header: /cvsroot/pgsql/src/backend/commands/async.c,v 1.91.2.1 2003/02/18 02:53:46 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -75,7 +75,6 @@
 #include <unistd.h>
 #include <signal.h>
 #include <errno.h>
-#include <sys/types.h>
 #include <netinet/in.h>
 
 #include "access/heapam.h"
@@ -85,6 +84,7 @@
 #include "libpq/libpq.h"
 #include "libpq/pqformat.h"
 #include "miscadmin.h"
+#include "storage/ipc.h"
 #include "tcop/tcopprot.h"
 #include "utils/fmgroids.h"
 #include "utils/ps_status.h"
@@ -150,7 +150,7 @@ void
 Async_Notify(char *relname)
 {
 	if (Trace_notify)
-		elog(DEBUG, "Async_Notify: %s", relname);
+		elog(LOG, "Async_Notify: %s", relname);
 
 	/* no point in making duplicate entries in the list ... */
 	if (!AsyncExistsPendingNotify(relname))
@@ -198,13 +198,13 @@ Async_Listen(char *relname, int pid)
 	bool		alreadyListener = false;
 
 	if (Trace_notify)
-		elog(DEBUG, "Async_Listen: %s", relname);
+		elog(LOG, "Async_Listen: %s", relname);
 
 	lRel = heap_openr(ListenerRelationName, AccessExclusiveLock);
 
 	/* Detect whether we are already listening on this relname */
-	scan = heap_beginscan(lRel, 0, SnapshotNow, 0, (ScanKey) NULL);
-	while (HeapTupleIsValid(tuple = heap_getnext(scan, 0)))
+	scan = heap_beginscan(lRel, SnapshotNow, 0, (ScanKey) NULL);
+	while ((tuple = heap_getnext(scan, ForwardScanDirection)) != NULL)
 	{
 		Form_pg_listener listener = (Form_pg_listener) GETSTRUCT(tuple);
 
@@ -221,7 +221,7 @@ Async_Listen(char *relname, int pid)
 	if (alreadyListener)
 	{
 		heap_close(lRel, AccessExclusiveLock);
-		elog(NOTICE, "Async_Listen: We are already listening on %s", relname);
+		elog(WARNING, "Async_Listen: We are already listening on %s", relname);
 		return;
 	}
 
@@ -241,17 +241,10 @@ Async_Listen(char *relname, int pid)
 	values[i++] = (Datum) 0;	/* no notifies pending */
 
 	tuple = heap_formtuple(RelationGetDescr(lRel), values, nulls);
-	heap_insert(lRel, tuple);
+	simple_heap_insert(lRel, tuple);
 
 #ifdef NOT_USED					/* currently there are no indexes */
-	if (RelationGetForm(lRel)->relhasindex)
-	{
-		Relation	idescs[Num_pg_listener_indices];
-
-		CatalogOpenIndices(Num_pg_listener_indices, Name_pg_listener_indices, idescs);
-		CatalogIndexInsert(idescs, Num_pg_listener_indices, lRel, tuple);
-		CatalogCloseIndices(Num_pg_listener_indices, idescs);
-	}
+	CatalogUpdateIndexes(lRel, tuple);
 #endif
 
 	heap_freetuple(tuple);
@@ -300,12 +293,12 @@ Async_Unlisten(char *relname, int pid)
 	}
 
 	if (Trace_notify)
-		elog(DEBUG, "Async_Unlisten %s", relname);
+		elog(LOG, "Async_Unlisten %s", relname);
 
 	lRel = heap_openr(ListenerRelationName, AccessExclusiveLock);
 
-	scan = heap_beginscan(lRel, 0, SnapshotNow, 0, (ScanKey) NULL);
-	while (HeapTupleIsValid(tuple = heap_getnext(scan, 0)))
+	scan = heap_beginscan(lRel, SnapshotNow, 0, (ScanKey) NULL);
+	while ((tuple = heap_getnext(scan, ForwardScanDirection)) != NULL)
 	{
 		Form_pg_listener listener = (Form_pg_listener) GETSTRUCT(tuple);
 
@@ -358,7 +351,7 @@ Async_UnlistenAll(void)
 	ScanKeyData key[1];
 
 	if (Trace_notify)
-		elog(DEBUG, "Async_UnlistenAll");
+		elog(LOG, "Async_UnlistenAll");
 
 	lRel = heap_openr(ListenerRelationName, AccessExclusiveLock);
 	tdesc = RelationGetDescr(lRel);
@@ -368,9 +361,9 @@ Async_UnlistenAll(void)
 						   Anum_pg_listener_pid,
 						   F_INT4EQ,
 						   Int32GetDatum(MyProcPid));
-	scan = heap_beginscan(lRel, 0, SnapshotNow, 1, key);
+	scan = heap_beginscan(lRel, SnapshotNow, 1, key);
 
-	while (HeapTupleIsValid(lTuple = heap_getnext(scan, 0)))
+	while ((lTuple = heap_getnext(scan, ForwardScanDirection)) != NULL)
 		simple_heap_delete(lRel, &lTuple->t_self);
 
 	heap_endscan(scan);
@@ -406,9 +399,9 @@ Async_UnlistenOnExit(void)
 	 */
 	AbortOutOfAnyTransaction();
 	/* Now we can do the unlisten */
-	StartTransactionCommand();
+	StartTransactionCommand(true);
 	Async_UnlistenAll();
-	CommitTransactionCommand();
+	CommitTransactionCommand(true);
 }
 
 /*
@@ -460,7 +453,7 @@ AtCommit_Notify(void)
 	}
 
 	if (Trace_notify)
-		elog(DEBUG, "AtCommit_Notify");
+		elog(LOG, "AtCommit_Notify");
 
 	/* preset data to update notify column to MyProcPid */
 	nulls[0] = nulls[1] = nulls[2] = ' ';
@@ -471,9 +464,9 @@ AtCommit_Notify(void)
 
 	lRel = heap_openr(ListenerRelationName, AccessExclusiveLock);
 	tdesc = RelationGetDescr(lRel);
-	scan = heap_beginscan(lRel, 0, SnapshotNow, 0, (ScanKey) NULL);
+	scan = heap_beginscan(lRel, SnapshotNow, 0, (ScanKey) NULL);
 
-	while (HeapTupleIsValid(lTuple = heap_getnext(scan, 0)))
+	while ((lTuple = heap_getnext(scan, ForwardScanDirection)) != NULL)
 	{
 		Form_pg_listener listener = (Form_pg_listener) GETSTRUCT(lTuple);
 		char	   *relname = NameStr(listener->relname);
@@ -492,14 +485,14 @@ AtCommit_Notify(void)
 			 */
 
 			if (Trace_notify)
-				elog(DEBUG, "AtCommit_Notify: notifying self");
+				elog(LOG, "AtCommit_Notify: notifying self");
 
 			NotifyMyFrontEnd(relname, listenerPID);
 		}
 		else
 		{
 			if (Trace_notify)
-				elog(DEBUG, "AtCommit_Notify: notifying pid %d",
+				elog(LOG, "AtCommit_Notify: notifying pid %d",
 					 listenerPID);
 
 			/*
@@ -528,14 +521,7 @@ AtCommit_Notify(void)
 				simple_heap_update(lRel, &lTuple->t_self, rTuple);
 
 #ifdef NOT_USED					/* currently there are no indexes */
-				if (RelationGetForm(lRel)->relhasindex)
-				{
-					Relation	idescs[Num_pg_listener_indices];
-
-					CatalogOpenIndices(Num_pg_listener_indices, Name_pg_listener_indices, idescs);
-					CatalogIndexInsert(idescs, Num_pg_listener_indices, lRel, rTuple);
-					CatalogCloseIndices(Num_pg_listener_indices, idescs);
-				}
+				CatalogUpdateIndexes(lRel, rTuple);
 #endif
 			}
 		}
@@ -555,7 +541,7 @@ AtCommit_Notify(void)
 	ClearPendingNotifies();
 
 	if (Trace_notify)
-		elog(DEBUG, "AtCommit_Notify: done");
+		elog(LOG, "AtCommit_Notify: done");
 }
 
 /*
@@ -613,6 +599,16 @@ Async_NotifyHandler(SIGNAL_ARGS)
 
 	if (notifyInterruptEnabled)
 	{
+		bool		save_ImmediateInterruptOK = ImmediateInterruptOK;
+
+		/*
+		 * We may be called while ImmediateInterruptOK is true; turn it off
+		 * while messing with the NOTIFY state.  (We would have to save
+		 * and restore it anyway, because PGSemaphore operations inside
+		 * ProcessIncomingNotify() might reset it.)
+		 */
+		ImmediateInterruptOK = false;
+
 		/*
 		 * I'm not sure whether some flavors of Unix might allow another
 		 * SIGUSR2 occurrence to recursively interrupt this routine. To
@@ -632,14 +628,21 @@ Async_NotifyHandler(SIGNAL_ARGS)
 			{
 				/* Here, it is finally safe to do stuff. */
 				if (Trace_notify)
-					elog(DEBUG, "Async_NotifyHandler: perform async notify");
+					elog(LOG, "Async_NotifyHandler: perform async notify");
 
 				ProcessIncomingNotify();
 
 				if (Trace_notify)
-					elog(DEBUG, "Async_NotifyHandler: done");
+					elog(LOG, "Async_NotifyHandler: done");
 			}
 		}
+
+		/*
+		 * Restore ImmediateInterruptOK, and check for interrupts if needed.
+		 */
+		ImmediateInterruptOK = save_ImmediateInterruptOK;
+		if (save_ImmediateInterruptOK)
+			CHECK_FOR_INTERRUPTS();
 	}
 	else
 	{
@@ -704,12 +707,12 @@ EnableNotifyInterrupt(void)
 		if (notifyInterruptOccurred)
 		{
 			if (Trace_notify)
-				elog(DEBUG, "EnableNotifyInterrupt: perform async notify");
+				elog(LOG, "EnableNotifyInterrupt: perform async notify");
 
 			ProcessIncomingNotify();
 
 			if (Trace_notify)
-				elog(DEBUG, "EnableNotifyInterrupt: done");
+				elog(LOG, "EnableNotifyInterrupt: done");
 		}
 	}
 }
@@ -760,13 +763,13 @@ ProcessIncomingNotify(void)
 				nulls[Natts_pg_listener];
 
 	if (Trace_notify)
-		elog(DEBUG, "ProcessIncomingNotify");
+		elog(LOG, "ProcessIncomingNotify");
 
 	set_ps_display("async_notify");
 
 	notifyInterruptOccurred = 0;
 
-	StartTransactionCommand();
+	StartTransactionCommand(true);
 
 	lRel = heap_openr(ListenerRelationName, AccessExclusiveLock);
 	tdesc = RelationGetDescr(lRel);
@@ -776,7 +779,7 @@ ProcessIncomingNotify(void)
 						   Anum_pg_listener_pid,
 						   F_INT4EQ,
 						   Int32GetDatum(MyProcPid));
-	scan = heap_beginscan(lRel, 0, SnapshotNow, 1, key);
+	scan = heap_beginscan(lRel, SnapshotNow, 1, key);
 
 	/* Prepare data for rewriting 0 into notification field */
 	nulls[0] = nulls[1] = nulls[2] = ' ';
@@ -785,7 +788,7 @@ ProcessIncomingNotify(void)
 	value[0] = value[1] = value[2] = (Datum) 0;
 	value[Anum_pg_listener_notify - 1] = Int32GetDatum(0);
 
-	while (HeapTupleIsValid(lTuple = heap_getnext(scan, 0)))
+	while ((lTuple = heap_getnext(scan, ForwardScanDirection)) != NULL)
 	{
 		Form_pg_listener listener = (Form_pg_listener) GETSTRUCT(lTuple);
 		char	   *relname = NameStr(listener->relname);
@@ -796,7 +799,7 @@ ProcessIncomingNotify(void)
 			/* Notify the frontend */
 
 			if (Trace_notify)
-				elog(DEBUG, "ProcessIncomingNotify: received %s from %d",
+				elog(LOG, "ProcessIncomingNotify: received %s from %d",
 					 relname, (int) sourcePID);
 
 			NotifyMyFrontEnd(relname, sourcePID);
@@ -805,14 +808,7 @@ ProcessIncomingNotify(void)
 			simple_heap_update(lRel, &lTuple->t_self, rTuple);
 
 #ifdef NOT_USED					/* currently there are no indexes */
-			if (RelationGetForm(lRel)->relhasindex)
-			{
-				Relation	idescs[Num_pg_listener_indices];
-
-				CatalogOpenIndices(Num_pg_listener_indices, Name_pg_listener_indices, idescs);
-				CatalogIndexInsert(idescs, Num_pg_listener_indices, lRel, rTuple);
-				CatalogCloseIndices(Num_pg_listener_indices, idescs);
-			}
+			CatalogUpdateIndexes(lRel, rTuple);
 #endif
 		}
 	}
@@ -827,7 +823,7 @@ ProcessIncomingNotify(void)
 	 */
 	heap_close(lRel, NoLock);
 
-	CommitTransactionCommand();
+	CommitTransactionCommand(true);
 
 	/*
 	 * Must flush the notify messages to ensure frontend gets them
@@ -838,7 +834,7 @@ ProcessIncomingNotify(void)
 	set_ps_display("idle");
 
 	if (Trace_notify)
-		elog(DEBUG, "ProcessIncomingNotify: done");
+		elog(LOG, "ProcessIncomingNotify: done");
 }
 
 /*
@@ -865,7 +861,7 @@ NotifyMyFrontEnd(char *relname, int32 listenerPID)
 		 */
 	}
 	else
-		elog(NOTICE, "NOTIFY for %s", relname);
+		elog(INFO, "NOTIFY for %s", relname);
 }
 
 /* Does pendingNotifies include the given relname? */
